// gcc eraser.c -o eraser.exe -lm -lpng -O3 -fast-math

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <float.h>
#include <math.h>
#include <png.h>
#include <windows.h>

#define GRANGE 7
#define GRAD 1.5f

typedef struct
{
#ifdef __APPLE__
	uint32_t a:8;
	uint32_t r:8;
	uint32_t g:8;
	uint32_t b:8;
#else
	uint32_t b:8;
	uint32_t g:8;
	uint32_t r:8;
	uint32_t a:8;
#endif
} rgb32_t;

uint8_t *load_png_file(const char *file, int32_t *width, int32_t *height, uint8_t mode)
{
	/* Modes :
	 * 0 : RGBX8888
	 * 1 : 8-bit grey
	 */

	FILE *fd;
	uint8_t *data, *image, header[8], ic;
	int32_t  bit_depth, color_type, i, iy, ix;
	png_uint_32  png_width, png_height, rowbytes;
	png_structp png_ptr;
	png_infop info_ptr;
	png_bytep *row_pointers;

	fd = fopen( file, "rb" );
	if ( fd == NULL )
	{
		perror(file);
		fprintf(stderr, "File %s missing\n", file);
		return NULL;
	}
	// ensure that we opened a PNG file
	fread( header, 1, 8, fd );
	if ( ! png_check_sig( header, 8 ) )
	{
		fclose(fd);
		printf("File %s does not have a valid PNG signature.\n", file);
		return NULL;
	}
	// create PNG structs
	png_ptr = png_create_read_struct( PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if ( ! png_ptr )
	{
		fclose(fd);
		return NULL;
	}
	info_ptr = png_create_info_struct(png_ptr);
	if ( ! info_ptr )
	{
		png_destroy_read_struct( &png_ptr, (png_infopp)NULL, (png_infopp)NULL);
		fclose(fd);
		return NULL;
	}
	//libpng does a longjmp here when it encounters an error
	if ( setjmp( png_ptr->jmpbuf ) )
	{
		png_destroy_read_struct( &png_ptr, &info_ptr, NULL);
		fclose(fd);
		return NULL;
	}
	png_init_io( png_ptr, fd );
	png_set_sig_bytes( png_ptr, 8);
	png_read_info( png_ptr, info_ptr);
	png_get_IHDR( png_ptr, info_ptr, &png_width, &png_height, &bit_depth,
	              &color_type, NULL, NULL, NULL);
	*width = (int32_t) png_width;
	*height = (int32_t) png_height;

//*********************SHOULDN'T BE NEEDED****************************
	//convert image format to RGB(A)
	if ( color_type == PNG_COLOR_TYPE_PALETTE )
		png_set_expand(png_ptr);

	if ((( color_type == PNG_COLOR_TYPE_GRAY ) || ( color_type == PNG_COLOR_TYPE_GRAY_ALPHA )) && mode!=3)
		png_set_gray_to_rgb(png_ptr);

	// we only handle 8 bits per channel
	if (bit_depth == 16 && mode!=3)
		png_set_strip_16(png_ptr);
	// insert filling bytes if necessary
	png_set_filler(png_ptr, 0xff, PNG_FILLER_AFTER);

	// update info data
	png_read_update_info( png_ptr, info_ptr);
//********************************************************************

	// allocate space for data and row pointers
	rowbytes = png_get_rowbytes( png_ptr, info_ptr);
	data = malloc (rowbytes * (*height));
	row_pointers = malloc ((*height) * sizeof(png_bytep));
	if (( data == NULL ) || ( row_pointers == NULL ))
	{
		png_destroy_read_struct( &png_ptr, &info_ptr, NULL);
		free(data);
		free(row_pointers);
		return NULL;
	}
	// initialize row pointers
	for ( i = 0;  i < *height; i++ )
		row_pointers[i] = data + i * rowbytes;
	// read all image data
	png_read_image( png_ptr, row_pointers );
	png_read_end( png_ptr, NULL);
	// cleanup
	free(row_pointers);
	png_destroy_read_struct( &png_ptr, &info_ptr, NULL);
	fclose(fd);
	// return an array of RGB(A) values
	//return data;

	if (mode==0)	//RGBX8888
	{
		image = malloc ((*width) * (*height) * 4 * sizeof(uint8_t));
		for (iy=0; iy<*height; iy++)
		{
			for (ix=0; ix<*width; ix++)
				for (ic=0; ic<3; ic++)
					image[(((iy*(*width))+ix)<<2)+ic] = data[(iy * (*width) + ix) * 4 + ic];
			Sleep(3);
		}
	}

	if (mode==1)	//8-bit grey
	{
		image = malloc ((*width) * (*height) * sizeof(uint8_t));
		for (iy=0; iy<*height; iy++)
		{
			for (ix=0; ix<*width; ix++)
				image[iy * (*width) + ix] =  data[(iy * (*width) + ix) * 4];
			Sleep(3);
		}
	}

	free(data);
	return image;
}

void fwrite_le_short(uint16_t s, FILE *file)	// write to file a 16-bit integer in little endian
{
	uint8_t byte;

	byte = s & 0xFF;
	fwrite(&byte, 1, 1, file);
	byte = (s>>8) & 0xFF;
	fwrite(&byte, 1, 1, file);
}

void fwrite_le_word(uint32_t w, FILE *file)	// write to file a 32-bit integer in little endian
{
	uint8_t byte;

	byte = w & 0xFF;
	fwrite(&byte, 1, 1, file);
	byte = (w>>8) & 0xFF;
	fwrite(&byte, 1, 1, file);
	byte = (w>>16) & 0xFF;
	fwrite(&byte, 1, 1, file);
	byte = (w>>24) & 0xFF;
	fwrite(&byte, 1, 1, file);
}

void save_image(rgb32_t *in, int32_t w, int32_t h, char *path)
{
	FILE *bmpfile = fopen(path, "wb");
	int32_t	i, iy, ix, ic;			// various iterators
	int32_t	filesize, imagesize;
	uint8_t	zerobytes, val, zero=0, r, g, b;

	zerobytes = 4 - ((w*3) & 3);		// computation of zero bytes
	if (zerobytes==4)
		zerobytes = 0;

	//********Tags********

	filesize = 56 + ((w*3)+zerobytes) * h;
	imagesize = 2 + ((w*3)+zerobytes) * h;

	fwrite_le_short(19778, bmpfile);
	fwrite_le_word(filesize, bmpfile);
	fwrite_le_word(0, bmpfile);
	fwrite_le_word(54, bmpfile);
	fwrite_le_word(40, bmpfile);
	fwrite_le_word(w, bmpfile);
	fwrite_le_word(h, bmpfile);
	fwrite_le_short(1, bmpfile);
	fwrite_le_word(24, bmpfile);
	fwrite_le_short(0, bmpfile);
	fwrite_le_word(imagesize, bmpfile);
	fwrite_le_word(2834, bmpfile);
	fwrite_le_word(2834, bmpfile);
	fwrite_le_word(0, bmpfile);
	fwrite_le_word(0, bmpfile);
	//--------Tags--------

	for (iy=h-1; iy>=0; iy--)		// backwards writing
	{
		for (ix=0; ix<w; ix++)
		{
			r = in[iy*w + ix].r;
			g = in[iy*w + ix].g;
			b = in[iy*w + ix].b;

			fwrite(&r, 1, 1, bmpfile);
			fwrite(&g, 1, 1, bmpfile);
			fwrite(&b, 1, 1, bmpfile);
		}

		for (i=0; i<zerobytes; i++)
			fwrite(&zero, 1, 1, bmpfile);	// write padding bytes

		Sleep(3);
	}

	fwrite_le_short(0, bmpfile);

	fclose(bmpfile);
}

float gaussian(float x)	// gaussian(x) = e^-x
{
	return expf(-x*x);
}

int main(int argc, char **argv)
{
	int32_t i, iy, ix, jy, jx, posx, posy, w, h;
	rgb32_t *image;//, *output;
	uint8_t *overlay;
	float lut[GRANGE*GRANGE], plut[GRANGE*GRANGE], sum, r, g, b;

	if (argc<4)
		fprintf(stderr, "Not enough arguments\neraser.exe image.png overlay.png output.bmp\n");

	overlay = load_png_file(argv[2], &w, &h, 1);
	image = (rgb32_t *) load_png_file(argv[1], &w, &h, 0);
	//output = calloc (w*h, sizeof(rgb32_t));
	//memcpy(output, image, w*h * sizeof(rgb32_t));

	for (iy=0; iy<GRANGE; iy++)
	{
		for (ix=0; ix<GRANGE; ix++)
		{
			lut[iy*GRANGE + ix] = gaussian((float) (GRANGE/2 - iy) / GRAD) * gaussian((float) (GRANGE/2 - ix) / GRAD);
		}
	}

	for (iy=0; iy<h; iy++)
	{
		for (ix=0; ix<w; ix++)
		{
			sum = 0.f;

			if (overlay[iy*w + ix] > 128)	// only interpolate pixels when overlay pixel > 128
			{
				// calculate pixel look up table
				for (jy=0; jy<GRANGE; jy++)
				{
					posy = iy + jy - GRANGE/2; if (posy<0) posy=0; if (posy>=h) posy=h-1;
	
					for (jx=0; jx<GRANGE; jx++)
					{
						posx = ix + jx - GRANGE/2; if (posx<0) posx=0; if (posx>=w) posx=w-1;
	
						if (overlay[posy*w + posx] == 0)	// if the pixel can be used as a source
							plut[jy*GRANGE + jx] = lut[jy*GRANGE + jx] ;
						else
							plut[jy*GRANGE + jx] = 0.f;
	
						sum += plut[jy*GRANGE + jx];
					}
				}
	
				if (sum==0.f)
					fprintf(stderr, "Error: zero sum for pixel %d , %d\n", ix, iy);
	
				r = g = b = 0.f;
	
				// calculate new pixel
				for (jy=0; jy<GRANGE; jy++)
				{
					posy = iy + jy - GRANGE/2; if (posy<0) posy=0; if (posy>=h) posy=h-1;
	
					for (jx=0; jx<GRANGE; jx++)
					{
						posx = ix + jx - GRANGE/2; if (posx<0) posx=0; if (posx>=w) posx=w-1;
	
						if (overlay[posy*w + posx] == 0)	// if the pixel can be used as a source
						{
							r += (float) image[posy*w + posx].r * plut[jy*GRANGE + jx] / sum;
							g += (float) image[posy*w + posx].g * plut[jy*GRANGE + jx] / sum;
							b += (float) image[posy*w + posx].b * plut[jy*GRANGE + jx] / sum;
						}
					}
				}
	
				image[iy*w + ix].r = (uint8_t) r + 0.5f;
				image[iy*w + ix].g = (uint8_t) g + 0.5f;
				image[iy*w + ix].b = (uint8_t) b + 0.5f;
			}
		}

		Sleep(3);
	}

	save_image(image, w, h, argv[3]);

	return 0;
}
